@@ -273,8 +273,7 @@ enum
 	PROP_SETTINGS,
 	PROP_HIGHLIGHT,
 	PROP_OCCURRENCES_COUNT,
-	PROP_REGEX_ERROR,
-	PROP_REGEX_STATE
+	PROP_REGEX_ERROR
 };
 
 struct _GtkSourceSearchContextPrivate
@@ -308,7 +307,6 @@ struct _GtkSourceSearchContextPrivate
 
 	GRegex *regex;
 	GError *regex_error;
-	GtkSourceRegexSearchState regex_state;
 
 	gint occurrences_count;
 	gulong idle_scan_id;
@@ -544,12 +542,6 @@ clear_search (GtkSourceSearchContext *search)
 		g_object_notify (G_OBJECT (search), "regex-error");
 	}
 
-	if (search->priv->regex_state != GTK_SOURCE_REGEX_SEARCH_NO_ERROR)
-	{
-		search->priv->regex_state = GTK_SOURCE_REGEX_SEARCH_NO_ERROR;
-		g_object_notify (G_OBJECT (search), "regex-state");
-	}
-
 	clear_task (search);
 
 	search->priv->occurrences_count = 0;
@@ -725,11 +717,7 @@ basic_forward_regex_search (GtkSourceSearchContext *search,
 
 	if (search->priv->regex_error != NULL)
 	{
-		search->priv->regex_state = GTK_SOURCE_REGEX_SEARCH_MATCHING_ERROR;
-
 		g_object_notify (G_OBJECT (search), "regex-error");
-		g_object_notify (G_OBJECT (search), "regex-state");
-
 		found = FALSE;
 	}
 
@@ -867,11 +855,7 @@ basic_backward_regex_search (GtkSourceSearchContext *search,
 
 	if (search->priv->regex_error != NULL)
 	{
-		search->priv->regex_state = GTK_SOURCE_REGEX_SEARCH_MATCHING_ERROR;
-
 		g_object_notify (G_OBJECT (search), "regex-error");
-		g_object_notify (G_OBJECT (search), "regex-state");
-
 		found = FALSE;
 	}
 
@@ -1355,8 +1339,9 @@ adjust_subregion (GtkSourceSearchContext *search,
 	});
 }
 
-/* Do not take into account the scan_region. Search just with
- * forward_to_tag_toggle(), and checks that it is not an old match.
+/* Do not take into account the scan_region. Take the result with a grain of
+ * salt. You should verify before or after calling this function that the
+ * region has been scanned, to be sure that the returned occurrence is correct.
  */
 static gboolean
 smart_forward_search_without_scanning (GtkSourceSearchContext *search,
@@ -1917,10 +1902,7 @@ regex_search_scan_segment (GtkSourceSearchContext *search,
 
 	if (search->priv->regex_error != NULL)
 	{
-		search->priv->regex_state = GTK_SOURCE_REGEX_SEARCH_MATCHING_ERROR;
-
 		g_object_notify (G_OBJECT (search), "regex-error");
-		g_object_notify (G_OBJECT (search), "regex-state");
 	}
 
 	if (g_match_info_is_partial_match (match_info))
@@ -2312,7 +2294,6 @@ static void
 update_regex (GtkSourceSearchContext *search)
 {
 	gboolean regex_error_changed = FALSE;
-	GtkSourceRegexSearchState new_regex_state = GTK_SOURCE_REGEX_SEARCH_NO_ERROR;
 	const gchar *search_text = gtk_source_search_settings_get_search_text (search->priv->settings);
 
 	if (search->priv->regex != NULL)
@@ -2353,7 +2334,6 @@ update_regex (GtkSourceSearchContext *search)
 
 		if (search->priv->regex_error != NULL)
 		{
-			new_regex_state = GTK_SOURCE_REGEX_SEARCH_COMPILATION_ERROR;
 			regex_error_changed = TRUE;
 		}
 
@@ -2363,12 +2343,6 @@ update_regex (GtkSourceSearchContext *search)
 		}
 	}
 
-	if (search->priv->regex_state != new_regex_state)
-	{
-		search->priv->regex_state = new_regex_state;
-		g_object_notify (G_OBJECT (search), "regex-state");
-	}
-
 	if (regex_error_changed)
 	{
 		g_object_notify (G_OBJECT (search), "regex-error");
@@ -2680,10 +2654,6 @@ gtk_source_search_context_get_property (GObject    *object,
 			g_value_set_pointer (value, gtk_source_search_context_get_regex_error (search));
 			break;
 
-		case PROP_REGEX_STATE:
-			g_value_set_enum (value, gtk_source_search_context_get_regex_state (search));
-			break;
-
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 			break;
@@ -2812,30 +2782,12 @@ gtk_source_search_context_class_init (GtkSourceSearchContextClass *klass)
 							       _("Regex error"),
 							       _("Regular expression error"),
 							       G_PARAM_READABLE));
-
-	/**
-	 * GtkSourceSearchContext:regex-state:
-	 *
-	 * The regex search state.
-	 *
-	 * Since: 3.10
-	 */
-	g_object_class_install_property (object_class,
-					 PROP_REGEX_STATE,
-					 g_param_spec_enum ("regex-state",
-							    _("Regex state"),
-							    _("State of the regular expression search"),
-							    GTK_SOURCE_TYPE_REGEX_SEARCH_STATE,
-							    GTK_SOURCE_REGEX_SEARCH_NO_ERROR,
-							    G_PARAM_READABLE));
 }
 
 static void
 gtk_source_search_context_init (GtkSourceSearchContext *search)
 {
 	search->priv = gtk_source_search_context_get_instance_private (search);
-
-	search->priv->regex_state = GTK_SOURCE_REGEX_SEARCH_NO_ERROR;
 }
 
 /**
@@ -3010,21 +2962,6 @@ gtk_source_search_context_get_regex_error (GtkSourceSearchContext *search)
 	return g_error_copy (search->priv->regex_error);
 }
 
-/**
- * gtk_source_search_context_get_regex_state:
- * @search: a #GtkSourceSearchContext.
- *
- * Returns: the regex search state.
- * Since: 3.10
- */
-GtkSourceRegexSearchState
-gtk_source_search_context_get_regex_state (GtkSourceSearchContext *search)
-{
-	g_return_val_if_fail (GTK_SOURCE_IS_SEARCH_CONTEXT (search), GTK_SOURCE_REGEX_SEARCH_NO_ERROR);
-
-	return search->priv->regex_state;
-}
-
 /**
  * gtk_source_search_context_get_occurrences_count:
  * @search: a #GtkSourceSearchContext.
@@ -3051,11 +2988,12 @@ gtk_source_search_context_get_occurrences_count (GtkSourceSearchContext *search)
  * @match_end: the end of the occurrence.
  *
  * Gets the position of a search occurrence. If the buffer is not already fully
- * scanned, the position may be unknown, and -1 is returned. Therefore you
- * should call this function when you know that the buffer is fully scanned.
+ * scanned, the position may be unknown, and -1 is returned. If 0 is returned,
+ * it means that this part of the buffer has already been scanned, and that
+ * @match_start and @match_end don't delimit an occurrence.
  *
  * Returns: the position of the search occurrence. The first occurrence has the
- * position 1 (not 0). Returns 0 if @match_start and @match_end doesn't delimit
+ * position 1 (not 0). Returns 0 if @match_start and @match_end don't delimit
  * an occurrence. Returns -1 if the position is not yet known.
  *
  * Since: 3.10
@@ -3070,6 +3008,8 @@ gtk_source_search_context_get_occurrence_position (GtkSourceSearchContext *searc
 	GtkTextIter iter;
 	gboolean found;
 	gint position = 0;
+	GtkTextRegion *region;
+	gboolean empty;
 
 	g_return_val_if_fail (GTK_SOURCE_IS_SEARCH_CONTEXT (search), -1);
 	g_return_val_if_fail (match_start != NULL, -1);
@@ -3080,6 +3020,27 @@ gtk_source_search_context_get_occurrence_position (GtkSourceSearchContext *searc
 		return -1;
 	}
 
+	/* Verify that the [match_start; match_end] region has been scanned. */
+
+	if (search->priv->scan_region != NULL)
+	{
+		region = gtk_text_region_intersect (search->priv->scan_region,
+						    match_start,
+						    match_end);
+
+		empty = is_text_region_empty (region);
+
+		if (region != NULL)
+		{
+			gtk_text_region_destroy (region, TRUE);
+		}
+
+		if (!empty)
+		{
+			return -1;
+		}
+	}
+
 	/* Verify that the occurrence is correct. */
 
 	found = smart_forward_search_without_scanning (search,
@@ -3103,11 +3064,11 @@ gtk_source_search_context_get_occurrence_position (GtkSourceSearchContext *searc
 
 	if (search->priv->scan_region != NULL)
 	{
-		GtkTextRegion *region = gtk_text_region_intersect (search->priv->scan_region,
-								   &iter,
-								   match_end);
+		region = gtk_text_region_intersect (search->priv->scan_region,
+						    &iter,
+						    match_end);
 
-		gboolean empty = is_text_region_empty (region);
+		empty = is_text_region_empty (region);
 
 		if (region != NULL)
 		{
@@ -3407,16 +3368,18 @@ gtk_source_search_context_backward_finish (GtkSourceSearchContext  *search,
 
 /* Returns %TRUE if replaced. */
 static gboolean
-regex_replace (GtkSourceSearchContext *search,
-	       GtkTextIter            *match_start,
-	       GtkTextIter            *match_end,
-	       const gchar            *replace)
+regex_replace (GtkSourceSearchContext  *search,
+	       GtkTextIter             *match_start,
+	       GtkTextIter             *match_end,
+	       const gchar             *replace,
+	       GError                 **error)
 {
 	GtkTextIter real_start;
 	gint start_pos;
 	gchar *subject;
 	gchar *subject_replaced;
 	GRegexMatchFlags match_options;
+	GError *tmp_error = NULL;
 
 	if (search->priv->regex == NULL ||
 	    search->priv->regex_error != NULL)
@@ -3436,17 +3399,13 @@ regex_replace (GtkSourceSearchContext *search,
 					    start_pos,
 					    replace,
 					    match_options,
-					    &search->priv->regex_error);
+					    &tmp_error);
 
 	g_free (subject);
 
-	if (search->priv->regex_error != NULL)
+	if (tmp_error != NULL)
 	{
-		search->priv->regex_state = GTK_SOURCE_REGEX_SEARCH_REPLACE_ERROR;
-
-		g_object_notify (G_OBJECT (search), "regex-error");
-		g_object_notify (G_OBJECT (search), "regex-state");
-
+		g_propagate_error (error, tmp_error);
 		g_free (subject_replaced);
 		return FALSE;
 	}
@@ -3460,21 +3419,6 @@ regex_replace (GtkSourceSearchContext *search,
 	return TRUE;
 }
 
-static void
-clear_replace_error (GtkSourceSearchContext *search)
-{
-	if (search->priv->regex_state == GTK_SOURCE_REGEX_SEARCH_REPLACE_ERROR)
-	{
-		search->priv->regex_state = GTK_SOURCE_REGEX_SEARCH_NO_ERROR;
-
-		g_error_free (search->priv->regex_error);
-		search->priv->regex_error = NULL;
-
-		g_object_notify (G_OBJECT (search), "regex-error");
-		g_object_notify (G_OBJECT (search), "regex-state");
-	}
-}
-
 /**
  * gtk_source_search_context_replace:
  * @search: a #GtkSourceSearchContext.
@@ -3482,6 +3426,7 @@ clear_replace_error (GtkSourceSearchContext *search)
  * @match_end: the end of the match to replace.
  * @replace: the replacement text.
  * @replace_length: the length of @replace in bytes, or -1.
+ * @error: location to a #GError, or %NULL to ignore errors.
  *
  * Replaces a search match by another text. If @match_start and @match_end
  * doesn't correspond to a search match, %FALSE is returned.
@@ -3494,11 +3439,12 @@ clear_replace_error (GtkSourceSearchContext *search)
  * Since: 3.10
  */
 gboolean
-gtk_source_search_context_replace (GtkSourceSearchContext *search,
-				   const GtkTextIter      *match_start,
-				   const GtkTextIter      *match_end,
-				   const gchar            *replace,
-				   gint                    replace_length)
+gtk_source_search_context_replace (GtkSourceSearchContext  *search,
+				   const GtkTextIter       *match_start,
+				   const GtkTextIter       *match_end,
+				   const gchar             *replace,
+				   gint                     replace_length,
+				   GError                 **error)
 {
 	GtkTextIter start;
 	GtkTextIter end;
@@ -3507,14 +3453,13 @@ gtk_source_search_context_replace (GtkSourceSearchContext *search,
 	g_return_val_if_fail (match_start != NULL, FALSE);
 	g_return_val_if_fail (match_end != NULL, FALSE);
 	g_return_val_if_fail (replace != NULL, FALSE);
+	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
 	if (dispose_has_run (search))
 	{
 		return FALSE;
 	}
 
-	clear_replace_error (search);
-
 	if (!smart_forward_search (search, match_start, &start, &end))
 	{
 		return FALSE;
@@ -3528,7 +3473,7 @@ gtk_source_search_context_replace (GtkSourceSearchContext *search,
 
 	if (gtk_source_search_settings_get_regex_enabled (search->priv->settings))
 	{
-		return regex_replace (search, &start, &end, replace);
+		return regex_replace (search, &start, &end, replace, error);
 	}
 
 	gtk_text_buffer_begin_user_action (search->priv->buffer);
@@ -3544,6 +3489,7 @@ gtk_source_search_context_replace (GtkSourceSearchContext *search,
  * @search: a #GtkSourceSearchContext.
  * @replace: the replacement text.
  * @replace_length: the length of @replace in bytes, or -1.
+ * @error: location to a #GError, or %NULL to ignore errors.
  *
  * Replaces all search matches by another text. It is a synchronous function, so
  * it can block the user interface.
@@ -3556,9 +3502,10 @@ gtk_source_search_context_replace (GtkSourceSearchContext *search,
  * Since: 3.10
  */
 guint
-gtk_source_search_context_replace_all (GtkSourceSearchContext *search,
-				       const gchar            *replace,
-				       gint                    replace_length)
+gtk_source_search_context_replace_all (GtkSourceSearchContext  *search,
+				       const gchar             *replace,
+				       gint                     replace_length,
+				       GError                 **error)
 {
 	GtkTextIter iter;
 	GtkTextIter match_start;
@@ -3569,16 +3516,17 @@ gtk_source_search_context_replace_all (GtkSourceSearchContext *search,
 
 	g_return_val_if_fail (GTK_SOURCE_IS_SEARCH_CONTEXT (search), 0);
 	g_return_val_if_fail (replace != NULL, 0);
+	g_return_val_if_fail (error == NULL || *error == NULL, 0);
 
 	if (dispose_has_run (search))
 	{
 		return 0;
 	}
 
-	clear_replace_error (search);
-
 	if (gtk_source_search_settings_get_regex_enabled (search->priv->settings))
 	{
+		GError *tmp_error = NULL;
+
 		if (search->priv->regex == NULL ||
 		    search->priv->regex_error != NULL)
 		{
@@ -3587,15 +3535,11 @@ gtk_source_search_context_replace_all (GtkSourceSearchContext *search,
 
 		g_regex_check_replacement (replace,
 					   &has_regex_references,
-					   &search->priv->regex_error);
+					   &tmp_error);
 
-		if (search->priv->regex_error != NULL)
+		if (tmp_error != NULL)
 		{
-			search->priv->regex_state = GTK_SOURCE_REGEX_SEARCH_REPLACE_ERROR;
-
-			g_object_notify (G_OBJECT (search), "regex-error");
-			g_object_notify (G_OBJECT (search), "regex-state");
-
+			g_propagate_error (error, tmp_error);
 			return 0;
 		}
 	}
@@ -3617,25 +3561,20 @@ gtk_source_search_context_replace_all (GtkSourceSearchContext *search,
 
 	while (smart_forward_search (search, &iter, &match_start, &match_end))
 	{
-		gboolean replaced = FALSE;
-
 		if (has_regex_references)
 		{
-			replaced = regex_replace (search, &match_start, &match_end, replace);
+			if (!regex_replace (search, &match_start, &match_end, replace, error))
+			{
+				break;
+			}
 		}
 		else
 		{
 			gtk_text_buffer_delete (search->priv->buffer, &match_start, &match_end);
 			gtk_text_buffer_insert (search->priv->buffer, &match_end, replace, replace_length);
-
-			replaced = TRUE;
-		}
-
-		if (replaced)
-		{
-			nb_matches_replaced++;
 		}
 
+		nb_matches_replaced++;
 		iter = match_end;
 	}
 
